// This may look like C code, but it is really -*- C++ -*-
// RTSS --- Railway Total System Simulator
// (c) TAKAGI Ryo (at Kogakuin University)
// SimpleEDLCModelCreator.cc
//  --- an instantiation of class SimpleEDLCModelCreator
// -----
// ChangeLog:
// 2007. 12. 7
//  File created.
// -----


#include "EnergyStorage.hh"

// -----
// Instantiate an object.
// オブジェクトをひとつインスタンス化する。
// -----
namespace
{
  SimpleEDLCModelCreator * x = new SimpleEDLCModelCreator ;
}
