// This may look like C code, but it is really -*- C++ -*-
// RTSS --- Railway Total System Simulator
// (c) TAKAGI Ryo (at Prof. SONE's Laboratory, the Univ. of Tokyo.)
// enumerat.hh --- header file that defines all enum types
// $Id: enumerat.hh,v 2.4 1994/12/27 12:03:47 rt Exp $

#ifndef ___ENUMERAT_HH
#define ___ENUMERAT_HH	// こうしておくと2度インクルードしても大丈夫！

enum car_type {Inverter};
enum car_stat {
    Power, Brake, Coast, Stn_stop, Const_vel, Regenerate, Between_stop
};
enum cs_command {None, Ec_auto, Ea_cvauto0, Ea_cvauto1,
		     Const_volt, Const_power, Torque_keisu};
enum dp_command {No_comm, Delay_yes, Buffer_yes};
enum feeder_shape {Long, Circle};
enum nextsta_pattern {
    Station, Station_pass, Pass_station, Pass_pass
};
enum stendpattern {Dim_0, Dim_1, Dim_2};
enum cs_mode {
    Inv_pmode1, Inv_pmode2, Inv_pmode3,
	Inv_bmode1, Inv_bmode2, Inv_bmode3,
	Torque_const, Power_const, Freerun	// 古いものを対応のため残す
    };

enum tresist_type {JNR_ec};

enum Next_To_Increment {
    Grad_Next, Curve_Next, Vlimit_Next, No_Next
};

enum control_disturbance_type {	// 列車群制御の種類
    Without_control, Stop_at_station, Slow_down_between
};

enum Startpoint_Increment {
    Grad_Start, Curve_Start, Vlimit_Start
};

enum pbchar_interpolation {	// powerdata, brakedata における補間法
    Linear, Inv_mode13
};

enum station_obj_type {		// station_obj の型
    Doubletrack_no_cross, Vanishing
};

#endif /* ! ___ENUMERAT_HH */
