// This may look like C code, but it is really -*- C++ -*-
// RTSS --- Railway Total System Simulator
// (c) TAKAGI Ryo (at Kogakuin University)
// readfile.cc --- read data from files
// -----
// ChangeLog:
// 2007. 10. 19
//  Modified the #include's to ISO C++ style. Removed the old comment lines
//  that preserved the section reading "T File". I do not think it is useful
//  anymore, now that the very powerful PCs are available...
// -----

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <string>
#include <cctype>
#include <iostream>

using std :: cerr ;
using std :: endl ;


#include "kidenrt.hh"


// Initialisation file read.
// ファイルリードと初期化
void
feeder :: readfile
()
{
  rf_gfile () ;
  rf_nfile () ;
  rf_hfile () ;
  rf_pfile () ;
  rf_sfile () ;
  rf_ffile () ;
}
