// -- This may look like C code, but it's really -*- C++ -*-
// RTSS --- Railway Total System Simulator
// (c) TAKAGI Ryo (at Kogakuin University)
// rtss_iconv.hh --- declaring iconv related functions
// -----
// ChangeLog:
// 2007. 11. 8
//  File created.
// -----

#include <cstdlib>

#ifdef SJIS_ICONV
#include <iconv.h>
#endif /* ! ___SJIS_ICONV */

#ifndef ___RTSS_ICONV_HH
#define ___RTSS_ICONV_HH	// こうしておくと2度インクルードしても大丈夫！

// -----
// Converting from UTF8 to SJIS.
// UTF8 から SJIS への変換。
// -----
void convertFromUTF8ToSJIS ( const char * , char * , size_t ) ;

#endif /* ! ___RTSS_ICONV_HH */
