// This may look like C code, but it is really -*- C++ -*-
// RTSS --- Railway Total System Simulator
// (c) TAKAGI Ryo (at Kogakuin University)
// OBESS_CurrentCapperCreator.cc
//  --- an instantiation of class OBESS_CurrentCapperCreator
// -----
// ChangeLog:
// 2010. 1. 26
//  File created.
// -----


#include "OBESS_CurrentCapper.hh"

// -----
// Instantiate an object.
// オブジェクトをひとつインスタンス化する。
// -----
namespace
{
  OBESS_CurrentCapperCreator * x
  = new OBESS_CurrentCapperCreator ;
}
