// This may look like C code, but it is really -*- C++ -*-
// RTSS --- Railway Total System Simulator
// (c) TAKAGI Ryo (at Kogakuin University)
// Ref_SOC_Curve_Container.hh --- declaration of class Ref_SOC_Curve_Container
// -----
// ChangeLog:
// 2010. 1. 28
//  File created.
// -----


#ifndef Ref_SOC_Curve_Container_HH
#define Ref_SOC_Curve_Container_HH

#include "Ref_SOC_Curve.hh"
#include "RS_XmlReader.hh"

class Ref_SOC_Curve_Container
  : public RS_ObjectContainer
{
public :
  bool create ( TiXmlNode const * & , vector < string > const & ,
                vector < RS_Object * > const & ) ;
  SOC_FeedForward_Data const * locate ( string const & ) const ;
  SOC_FeedForward_Data * locate ( string const & ) ;
  ~Ref_SOC_Curve_Container () {}
} ;


#endif  // Ref_SOC_Curve_Container_HH
