// This may look like C code, but it is really -*- C++ -*-
// RTSS --- Railway Total System Simulator
// (c) TAKAGI Ryo (at Kogakuin University)
// Ref_SOC_Curve_Creator.cc --- instantiator of class Ref_SOC_Curve
// -----
// ChangeLog:
// 2010. 1. 28
//  File created.
// -----


#include "Ref_SOC_Curve.hh"

// -----
// Instantiate an object.
// オブジェクトをひとつインスタンス化する。
// -----
namespace
{
  Ref_SOC_Curve_Creator * x = new Ref_SOC_Curve_Creator ;
}
